#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#------------------------------------------------------------------------------------------------
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# -----------------------------------------------------------------------------------------------
# THIS PLUGIN WAS DEVELOPED AND TESTED UNDER WINDOWS VISTA ONLY AND MAY OR MAY NOT WORK ON A MAC. 
# -----------------------------------------------------------------------------------------------
# THIS PLUGIN IS ANOTHER FIGMENT OF MY IMAGENATION AND IS NOT BASED ON OR COMPLY WITH ANYTHING! 
# -----------------------------------------------------------------------------------------------
#		Name:	PolyLine
#	 Usage:	Interactive Freehand Curve Generator.
#		Note:	Pressing the Backspace key will delete the last point picked, the Alt key ends the curve 
#					at the last point, the Ctrl key closes the curve and double click ends at current cursor
#					location.  Entering a number for intermediate points smooths the curve by a spline
#					interpolation.  Entering a length with an 's' prefix or suffix sets the default segment 
#					length while 'dragging'.
#		Date:	Jan 2013
# -----------------------------------------------------------------------------------------------
#
require 'sketchup'
#
module SDM

	class PolyLine

		@@segment_length=nil; @@ipts=nil
		
		def initialize
			@@segment_length=1.feet unless @@segment_length
			@ip0 = Sketchup::InputPoint.new
			@ip1 = Sketchup::InputPoint.new	
			@vue = Sketchup.active_model.active_view
			@total=nil;@@ipts = 5 unless @@ipts
			@btn_dwn=false
			self.reset
		end
		
		def reset
			@total ? @last_total=@total : @last_total=0
			@pts = [];@state = 0;@total = 0;@seg = 0;@ip1.clear;@drawn = false
			@status_text = "Select 1st point of new curve.  #{@@ipts} intermediate points will be added.  " \
				+ "Default Segment Length=#{Sketchup.format_length(@@segment_length)}.  Last curve length=#{Sketchup.format_length(@last_total)}"
			Sketchup::set_status_text "", SB_VCB_LABEL
			Sketchup::set_status_text "", SB_VCB_VALUE
		end

		def onMouseMove(flags, x, y, view)
			self.set_current_point(x, y, view)
			if @state > 0
				@status_text = "Select point #{@state+1}, Backspace=Delete last point, Alt=End at last point, Ctrl=Close loop, " + \
				"Intermeadiate points #{@@ipts}, Segment length #{@@segment_length}"
			end
			Sketchup::set_status_text @status_text
			@vue.refresh
		end

		def onLButtonDown(flags, x, y, view)
			self.set_current_point(x, y, view)
			@btn_dwn=true;@state += 1; @total += @seg;@ip1.copy! @ip0
		end

		def onLButtonUp(flags, x, y, view)
			@btn_dwn=false
		end
		
		def onLButtonDoubleClick(flags, x, y, view)
			self.create_spline if @pts.length > 2
			self.reset
		end
		
		def set_current_point(x, y, view)
			@ip0.pick view,x,y
			@pts[@state] = @ip0.position
			view.tooltip = @ip0.tooltip	
			if @state > 0
				@seg=@pts[@state].distance(@pts[@state-1]);tot=@total+@seg
				Sketchup::set_status_text "Segment,Total:", SB_VCB_LABEL
				Sketchup::set_status_text "#{Sketchup.format_length(@seg)},#{Sketchup.format_length(tot)}", SB_VCB_VALUE
				@state += 1 if (@btn_dwn && (@pts[@state-1].distance(@pts[@state]) >= @@segment_length))
			end
		end
		
		def onRButtonDown(flags, x, y, view)
			if @state == 0
				Sketchup.send_action "selectSelectionTool:" 
			else
				self.create_spline if @pts.length > 2
				self.reset
			end
		end
		
		def onKeyDown(key,rpt,flags,view)
			if @pts.length > 2
				case key
					when  8 # Backspace key = remove last point
						(@pts.pop;@state -= 1;view.refresh) if @state > 0
					when VK_CONTROL # 17 # Ctrl key = close spline
						@pts[@state] = @pts[0]
						self.create_spline
						self.reset
					when VK_ALT # 18 # Alt key = end spline at previous point
						(@pts.pop;self.create_spline) if @state > 2
						self.reset
					when VK_SHIFT # 16 # Shift key = end at current location
						self.create_spline if @state > 2
						self.reset
				end#case
			end#if
		end#def
		
		def onCancel(flag, view)
			view.invalidate if @drawn
			Sketchup.send_action "selectSelectionTool:"
		end

		def onUserText(text, view)
			begin
				if text.include?('s')
					text.delete!('s')
					sl=text.to_l;sl=[1,sl].max
					@@segment_length=sl
				else
					ip=text.to_i
					ip=[0,ip].max;ip=[ip,9].min
					@@ipts=ip
				end
				if @state == 0
					@status_text = "Select 1st point of new curve.  #{@@ipts} intermediate points will be added.  " \
						+ "Default Segment Length=#{Sketchup.format_length(@@segment_length)}"
				else
					@status_text = "Select point #{@state+1}, Backspace=Delete last point, Alt=End at last point, Ctrl=Close loop, " + \
					"Intermeadiate points #{@@ipts}"
				end
				view.refresh
			rescue
				UI.beep
			end
		end

		def draw(view)
			if( @ip0.valid? && @ip0.display? )
				@ip0.draw(view)
			end
			case @state
				when 1
					view.line_width = 2
					view.drawing_color = "magenta"
					inference_locked = view.inference_locked?
					view.draw(GL_LINE_STRIP,@pts[0],@ip0.position)
				when 2..9999
					view.line_width = 2
					view.drawing_color = "magenta"
					inference_locked = view.inference_locked?
					if @@ipts > 0
						spts = self.get_spts(0)
						view.draw(GL_LINE_STRIP,spts)
					else
						view.draw(GL_LINE_STRIP,@pts)
					end
			end#case
		end
		
		def get_spts(k)
			inc = 1.0/@@ipts
			x = []; y = []; z = []; spts = []
			pts = self.add_pts
			self.get_xyz(pts,x,y,z)
			for i in (1..@np-2)
				inc.step(1.0,inc) do |mu|
					xs = self.sdm_cat_rom_interpolation(x[i-1],x[i],x[i+1],x[i+2],mu)
					ys = self.sdm_cat_rom_interpolation(y[i-1],y[i],y[i+1],y[i+2],mu)
					zs = self.sdm_cat_rom_interpolation(z[i-1],z[i],z[i+1],z[i+2],mu)
					spts[k] = [xs,ys,zs]
					k += 1
				end#step
			end#for i
			return spts
		end

		def sdm_cat_rom_interpolation(p0,p1,p2,p3,mu)
			a0 = -0.5*p0 + 1.5*p1 - 1.5*p2 + 0.5*p3;
			a1 = p0 - 2.5*p1 + 2*p2 - 0.5*p3;
			a2 = -0.5*p0 + 0.5*p2;
			a3 = p1;
			return (a0*mu**3+a1*mu**2+a2*mu+a3)
		end
		
		def create_spline
			Sketchup.active_model.start_operation "Create Spline"
			if @@ipts > 0
				spts = self.get_spts(1)
				spts[0] = @pts[0]
				spts.push @pts[-1]
				curve = Sketchup.active_model.entities.add_curve(spts)
			else
				curve = Sketchup.active_model.entities.add_curve(@pts)
			end
			@total=0; curve.each{|e| @total += e.length}
			Sketchup.active_model.commit_operation
		end

		def add_pts
			pts = []
			@pts.each {|e| pts.push e}
			if pts[0] == pts[-1] then #closed loop
				pts.push pts[1]
				pts.insert(0,pts[-2])
			else
				pt=pts[0].offset(pts[1].vector_to(pts[0]),10)
				pts.insert(0,pt)
				pt=pts[-1].offset(pts[-2].vector_to(pts[-1]),10)
				pts.push pt
			end
			@np = pts.length-1
			return pts
		end

		def get_xyz(pts,x,y,z)
			for i in (0..@np)
				x[i] = pts[i].x
				y[i] = pts[i].y
				z[i] = pts[i].z
			end
		end

	end#of class

end#of module
# ------------------  MENU SETUP ---------------------- #
unless $sdm_tools_menu
	$sdm_tools_menu = UI.menu("Plugins").add_submenu("Werkzeuge von sdmitch")
	$sdm_Edge_tools = $sdm_tools_menu.add_submenu("Kanten und Linien")
	$sdm_Face_tools = $sdm_tools_menu.add_submenu("Flächen Werkzeuge")
	$sdm_CorG_tools = $sdm_tools_menu.add_submenu("Gruppen und Komponenten")
	$sdm_Misc_tools = $sdm_tools_menu.add_submenu("Diverse Werkzeuge")
end
unless file_loaded?(__FILE__)
	$sdm_Edge_tools.add_item("Linienzug") { Sketchup.active_model.select_tool SDM::PolyLine.new }
	file_loaded(__FILE__)
end
# ------------------------------------------------------ #
